(* 
 * Copyright 2012 Damien Guichard  
 *  
 * Licensed under the EUPL, Version 1.1 only (the "Licence");
 * You may not use this work except in compliance with the Licence. 
 * You may obtain a copy of the Licence at: 
 *  
 * http://ec.europa.eu/idabc/eupl5  
 *  
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the Licence for the specific language governing permissions
 * and limitations under the Licence. 
 *)

module type TaxonomyTree
=
sig
	include ERicTypes.TaxonomyTree
	type partial_order = Less | Equal | More | Left | Right 
	val compare : t -> t -> partial_order 
end 

module TaxonomyTree
:
TaxonomyTree
=
struct
	type t =
		{
		alt : t option;
		mutable sub : t option;
		str : string;
		mutable min : int;
		mutable max : int;
		mutable cnt : int;
		}
	type set =
		t
	type partial_order =
		Less | Equal | More | Left | Right
	let singleton str =
		let t = {alt=None;sub=None;str;min=1;max=1;cnt=0}
		in t,t
	let from_string s set =
		let result = ref set in
		let rec loop = function
			| None -> ()
			| Some t ->
				loop t.sub;
				if t.str = s then (result := t; raise Exit);
				loop t.alt
		in
		try
			if set.str = s then raise Exit;
			loop set.sub; raise Not_found;
		with
		| Exit -> !result 
	let to_string t =
		t.str 
	let unique t =
		t.cnt <- pred t.cnt; t.cnt
	let compare ta tb = 
		if ta.max < tb.min then Left
		else if ta.min > tb.max then Right
		else if ta.min > tb.min && ta.max <= tb.max then Less
		else if tb.min > ta.min && tb.max <= ta.max then More
		else Equal 
	let rec renumber m = function
		| None -> m
		| Some t ->
			t.min <- m+1;
			t.max <- renumber (m+1) t.sub;
			renumber t.max t.alt
	let renumber t =
		ignore(renumber 0 (Some t))
	let add_childs_to childs t set =
		let sub = t.sub in
		let rec loop = function
			| [] -> ()
			| str::tl ->
				try
					ignore(from_string str set);
					t.sub <- sub;
					failwith "identifier already in taxonomy"
				with Not_found ->
					t.sub <- Some {alt=t.sub;sub=None;str;min=0;max=0;cnt=0};
					loop tl
		in loop childs; renumber set
end

module Graph (Taxo:TaxonomyTree)
:
ERicTypes.Graph with type label=Taxo.t
=
struct
	type label =
		Taxo.t
	type entity =
		| EntityL of label
		| EntityI of label * int
		| EntityR of label * int
		| EntityN of label * string
		| EntityS of label * string
		| EntityF of label * float
	type relation =
		label
	type edge =
		{
		mutable src : entity; src_mark : string;
		mutable rel : relation;
		mutable dst : entity; dst_mark : string;
		}
	type vertex =
		| Vertex
		| Source of edge
		| Destination of edge
	type pattern =
		| Nil
		| Pat of edge * vertex * vertex * pattern
		| Not of edge * vertex * vertex * pattern
		| Neq of entity * entity * vertex * vertex * pattern
	type t =
		graph option
	and graph =
		{
		mutable left  : t;
		mutable right : t;
		mutable other : t;
		t_src : entity;
		t_rel : relation;
		t_dst : entity;
		}
	let label = function
		| EntityL l
		| EntityI (l,_)
		| EntityR (l,_) 
		| EntityN (l,_)
		| EntityS (l,_)
		| EntityF (l,_) -> l 
	let label_compare a b =
		Taxo.compare (label a) (label b)
	let empty =
		None
	let singleton e =
		Some
		{left=None;right=None;other=None;t_src=e.src;t_rel=e.rel;t_dst=e.dst}
	let not_equal a b =
		a.src <> b.t_src or 
		a.dst <> b.t_dst or
		Taxo.compare a.rel b.t_rel <> Taxo.Equal
	let insert iter e g =
		let open Taxo in function
		| Left -> g.left <- iter e g.left
		| Right -> g.right <- iter e g.right
		| Less | More -> g.other <- iter e g.other
		| Equal -> if not_equal e g then g.other <- iter e g.other
	let rec insert1 e t =
		match t with
		| None -> singleton e
		| Some g -> insert insert2 e g (label_compare e.src g.t_src); t
	and insert2 e t =
		match t with
		| None -> singleton e
		| Some g -> insert insert3 e g (Taxo.compare e.rel g.t_rel); t
	and insert3 e t =
		match t with
		| None -> singleton e
		| Some g -> insert insert1 e g (label_compare e.dst g.t_dst); t
	let rec insert l t =
		match l with
		| [] -> t
		| e::tl -> insert tl (insert1 e t)
	let entity_compare ea eb =
		let open Taxo in
		match ea,eb with
		| _ when ea = eb -> Equal
		| _,EntityL l when compare (label ea) l < More -> Less 
		| EntityI(la,a),EntityI(lb,b)
		| EntityR(la,a),EntityR(lb,b) when a=b -> compare la lb
		| EntityN(la,a),EntityN(lb,b)
		| EntityS(la,a),EntityS(lb,b) when a=b -> compare la lb
		| EntityF(la,a),EntityF(lb,b) when a=b -> compare la lb
		| _ -> More
	let less_less a b =
		let open Taxo in	
		entity_compare a.t_src b.src < More &&
		entity_compare a.t_dst b.dst < More &&
		Taxo.compare   a.t_rel b.rel < More
	let equal_equal a b =
		let open Taxo in		
		entity_compare a.t_src b.src = Equal &&
		entity_compare a.t_dst b.dst = Equal &&
		Taxo.compare   a.t_rel b.rel < More
	let less_equal a b =
		let open Taxo in		
		entity_compare a.t_src b.src < More  &&
		entity_compare a.t_dst b.dst = Equal &&
		Taxo.compare   a.t_rel b.rel < More
	let equal_less a b =
		let open Taxo in		
		entity_compare a.t_src b.src = Equal &&
		entity_compare a.t_dst b.dst < More  &&
		Taxo.compare   a.t_rel b.rel < More
	let iterate cmp e f t =
		let iterate iter g =
			let open Taxo in function
			| Left -> iter g.left
			| Right -> iter g.right
			| Less -> iter g.other
			| Equal ->
				if cmp g e then f g; iter g.other
			| More ->
				if cmp g e then f g;
				iter g.left; iter g.other; iter g.right
		in			
		let rec iterate1 = function
			| None -> ()
			| Some g -> iterate iterate2 g (label_compare e.src g.t_src)
		and iterate2 = function
			| None -> ()
			| Some g -> iterate iterate3 g (Taxo.compare e.rel g.t_rel)
		and iterate3 = function
			| None -> ()
			| Some g -> iterate iterate1 g (label_compare e.dst g.t_dst)
		in iterate1 t
	let assign def = function
		| Vertex -> def
		| Source e -> e.src
		| Destination e -> e.dst
	let nothing () =
		()
	let printer =
		ref nothing
	let select_compare va vb =
		match va,vb with
		| Vertex,Vertex -> less_less
		| Vertex,_ -> less_equal
		| _,Vertex -> equal_less
		| _ -> equal_equal
	let rec select pat t =
		let range tl a b =
			a.src <- b.t_src;
			a.dst <- b.t_dst;
			a.rel <- b.t_rel;
			select tl t
		and stop _ =
			raise Exit
		in match pat with
		| Nil ->
			!printer ()
		| Pat(e,va,vb,tl) ->
			let triple =
				{e with src = assign e.src va; dst = assign e.dst vb}
			in
				iterate (select_compare va vb) triple (range tl e) t;
				e.src <- triple.src;
				e.dst <- triple.dst;
				e.rel <- triple.rel
		| Neq(ea,eb,va,vb,tl) ->
			let a = assign ea va and b = assign eb vb in
			( match a,b with
			| EntityL _,_ | _,EntityL _ -> select tl t 
			| _ -> if a <> b then select tl t)
		| Not(e,va,vb,tl) ->
			let triple =
				{e with src = assign e.src va; dst = assign e.dst vb}
			in
			try
				iterate (select_compare va vb) triple stop t;
				select tl t
			with Exit -> ()
	let select print pat t =
		printer := print; select pat t; printer := nothing
	let marshal_flags =
		[] 
end
